import pytest

from pytest_mock_resources.compat import pymongo
from pytest_mock_resources.config import DockerContainerConfig, fallback
from pytest_mock_resources.container.base import ContainerCheckFailed, get_container


class MongoConfig(DockerContainerConfig):
    """Define the configuration object for mongo.

    Args:
        image (str): The docker image:tag specifier to use for mongo containers.
            Defaults to :code:`"mongo:3.6"`.
        host (str): The hostname under which a mounted port will be available.
            Defaults to :code:`"localhost"`.
        port (int): The port to bind the container to.
            Defaults to :code:`28017`.
        ci_port (int): The port to bind the container to when a CI environment is detected.
            Defaults to :code:`27017`.
        root_database (str): The name of the root mongo database to create.
            Defaults to :code:`"dev-mongo"`.
    """

    name = "mongo"

    _fields = {"image", "host", "port", "ci_port", "root_database"}
    _fields_defaults = {
        "image": "mongo:3.6",
        "port": 28017,
        "ci_port": 27017,
        "root_database": "dev-mongo",
    }

    @fallback
    def root_database(self):
        raise NotImplementedError()


def get_pymongo_client(config):
    uri = "mongodb://{}:{}".format(config.host, config.port)

    return pymongo.MongoClient(uri)


def check_mongo_fn(config):
    try:
        client = get_pymongo_client(config)
        db = client[config.root_database]
        db.command("ismaster")
    except pymongo.errors.ConnectionFailure:
        raise ContainerCheckFailed(
            "Unable to connect to a presumed MongoDB test container via given config: {}".format(
                config
            )
        )


@pytest.fixture(scope="session")
def _mongo_container(pmr_mongo_config):
    result = get_container(pmr_mongo_config, {27017: pmr_mongo_config.port}, {}, check_mongo_fn)
    yield next(iter(result))



import pytest

from pytest_mock_resources.compat import pymongo
from pytest_mock_resources.container.mongo import get_pymongo_client, MongoConfig
from pytest_mock_resources.fixture.database.generic import assign_fixture_credentials


@pytest.fixture(scope="session")
def pmr_mongo_config():
    """Override this fixture with a :class:`MongoConfig` instance to specify different defaults.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_mongo_config():
        ...     return MongoConfig(image="mongo:3.4", root_database="foo")
    """
    return MongoConfig()


def create_mongo_fixture(**kwargs):
    scope = kwargs.pop("scope", "function")

    if len(kwargs):
        raise KeyError("Unsupported Arguments: {}".format(kwargs))

    @pytest.fixture(scope=scope)
    def _(_mongo_container, pmr_mongo_config):
        return _create_clean_database(pmr_mongo_config)

    return _


def _create_clean_database(config):
    client = get_pymongo_client(config)
    db = client[config.root_database]

    # Create a collection called `pytestMockResourceDbs' in the admin tab if it hasnt already been
    # created.
    db_collection = db["pytestMockResourcesDbs"]

    #  create a Document in the `pytestMockResourcesDbs` collection:
    result = db_collection.insert_one({})
    db_id = str(result.inserted_id)

    #  Create a database where the name is equal to that ID:
    create_db = client[db_id]

    #  Create a user as that databases owner
    create_db.command("createUser", db_id, pwd="password", roles=["dbOwner"])  # nosec

    #  pass back an authenticated db connection
    limited_client = pymongo.MongoClient(config.host, config.port)
    db = limited_client[db_id]
    db.authenticate(db_id, "password")

    db.config = {"username": db_id, "password": "password", "database": db_id}

    assign_fixture_credentials(
        db,
        drivername="mongodb",
        host=config.host,
        port=config.port,
        database=db_id,
        username=db_id,
        password="password",
    )

    return db



_resource_kinds = ["postgres", "redshift", "mongo", "redis", "mysql"]


def pytest_configure(config):
    """Register markers for each resource kind."""
    for resource_kind in _resource_kinds:
        config.addinivalue_line(
            "markers",
            "postgres: Tests which make use of {kind} fixtures".format(kind=resource_kind),
        )


def pytest_itemcollected(item):
    """Attach markers to each test which uses a fixture of one of the resources."""
    if not hasattr(item, "fixturenames"):
        return

    fixturenames = set(item.fixturenames)
    for resource_kind in _resource_kinds:
        resource_fixture = "_{}_container".format(resource_kind)
        if resource_fixture in fixturenames:
            item.add_marker(resource_kind)



import pytest
import sqlalchemy
from sqlalchemy.engine.url import URL

from pytest_mock_resources.config import DockerContainerConfig, fallback
from pytest_mock_resources.container.base import ContainerCheckFailed, get_container


class MysqlConfig(DockerContainerConfig):
    """Define the configuration object for MySql.

    Args:
        image (str): The docker image:tag specifier to use for postgres containers.
            Defaults to :code:`"postgres:9.6.10-alpine"`.
        host (str): The hostname under which a mounted port will be available.
            Defaults to :code:`"localhost"`.
        port (int): The port to bind the container to.
            Defaults to :code:`5532`.
        ci_port (int): The port to bind the container to when a CI environment is detected.
            Defaults to :code:`5432`.
        username (str): The username of the root postgres user
            Defaults to :code:`"user"`.
        password (str): The password of the root postgres password
            Defaults to :code:`"password"`.
        root_database (str): The name of the root postgres database to create.
            Defaults to :code:`"dev"`.
    """

    name = "mysql"
    _fields = {"image", "host", "port", "ci_port", "username", "password", "root_database"}
    _fields_defaults = {
        "image": "mysql:5.6",
        "port": 3406,
        "ci_port": 3306,
        "username": "root",
        "password": "password",
        "root_database": "dev",
    }

    @fallback
    def username(self):
        raise NotImplementedError()

    @fallback
    def password(self):
        raise NotImplementedError()

    @fallback
    def root_database(self):
        raise NotImplementedError()


def get_sqlalchemy_engine(config, database_name, isolation_level=None):
    DB_URI = str(
        URL(
            "mysql+pymysql",
            username=config.username,
            password=config.password,
            host=config.host,
            port=config.port,
            database=database_name,
        )
    )

    options = {}
    if isolation_level:
        options["isolation_level"] = isolation_level

    from pytest_mock_resources.compat import pymysql

    pymysql.connect

    engine = sqlalchemy.create_engine(DB_URI, **options)

    # Verify engine is connected
    engine.connect()

    return engine


def check_mysql_fn(config):
    try:
        get_sqlalchemy_engine(config, config.root_database)
    except sqlalchemy.exc.OperationalError:
        raise ContainerCheckFailed(
            "Unable to connect to a presumed MySQL test container via given config: {}".format(
                config
            )
        )


@pytest.fixture(scope="session")
def _mysql_container(pmr_mysql_config):
    result = get_container(
        pmr_mysql_config,
        {3306: pmr_mysql_config.port},
        {
            "MYSQL_DATABASE": pmr_mysql_config.root_database,
            "POSTGRES_USER": pmr_mysql_config.username,
            "MYSQL_ROOT_PASSWORD": pmr_mysql_config.password,
        },
        check_mysql_fn,
    )

    yield next(iter(result))



"""Adapt SQLite to act more in line with typical postgresql usage.

SQLite is a desireable test database target because it introduces much less latency across
tests, is parallelizable, and has a lower minimum fixed test-startup cost.

Unfortunately, SQLite and postgresql do not behave identically in all cases. However, in
many (typically CRUD) API type database designs, there is no need for postgresql-specific
features, and SQLite can act as an appropriate stand in.

Where possible though, this module changes the default SQLite behavior to more closely
mimic postgresql, so as to increase the number of places where SQLite can frictionlessly
stand in for postgresql.
"""
import contextlib
import datetime
import json
import warnings

import pytest
from sqlalchemy import create_engine, event
from sqlalchemy.dialects import registry
from sqlalchemy.dialects.postgresql import JSON, JSONB
from sqlalchemy.dialects.sqlite import base as sqlite_base
from sqlalchemy.dialects.sqlite.pysqlite import SQLiteDialect_pysqlite
from sqlalchemy.exc import SAWarning
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import sqltypes

from pytest_mock_resources.fixture.database.generic import assign_fixture_credentials
from pytest_mock_resources.fixture.database.relational.generic import EngineManager


class PMRSQLiteDDLCompiler(sqlite_base.SQLiteDDLCompiler):
    """Add features sqlite doesn't support by default.

    Adds:
     * Handling `CREATE SCHEMA`.
    """

    def visit_create_schema(self, create):
        schema = self.preparer.format_schema(create.element)
        return "ATTACH DATABASE ':memory:' AS " + schema

    def visit_drop_schema(self, create):
        schema = self.preparer.format_schema(create.element)
        return "DETACH DATABASE " + schema


class UTC(datetime.tzinfo):
    """UTC timezone.

    Blame python2:
    https://docs.python.org/2/library/datetime.html#datetime.tzinfo.fromutc
    """

    ZERO = datetime.timedelta(0)

    def utcoffset(self, dt):
        return self.ZERO

    def dst(self, dt):
        return self.ZERO

    def tzname(self, dt):
        return "UTC"


utc = UTC()


class PMRDateTime(sqlite_base.DATETIME):
    def bind_processor(self, dialect):
        processor = super(PMRDateTime, self).bind_processor(dialect)

        def process(value):
            if isinstance(value, datetime.datetime):
                if value.tzinfo is None:
                    value = value.replace(tzinfo=utc)
                result = value.astimezone(utc)
                return processor(result)

        return process

    def result_processor(self, dialect, coltype):
        processor = super(PMRDateTime, self).result_processor(dialect, coltype)

        def process(value):
            result = processor(value)
            if self.timezone:
                if result is not None and result.tzinfo is None:
                    result = result.replace(tzinfo=utc)
            return result

        return process


PostgresLikeSQLitePDialect = None


def make_postgres_like_sqlite_dialect():
    if not hasattr(sqlite_base, "JSON"):
        raise RuntimeError(
            "Must have sqlalchemy>=1.3.0 in order to make use of the postgres-like sqlite dialect."
        )

    class PostgresLikeTypeCompiler(sqlite_base.SQLiteTypeCompiler):
        """Add features sqlite doesn't support by default.

        Adds:
         * Handling of postgres column types: JSON, JSONB
        """

        @compiles(JSON, "pmrsqlite")
        @compiles(JSONB, "pmrsqlite")
        @compiles(sqltypes.JSON, "pmrsqlite")
        def compile_json(type_, compiler, **kwargs):
            return "BLOB"

    class PostgresLikeDialect(SQLiteDialect_pysqlite):
        """Define the dialect that collects all postgres-like adapations."""

        name = "pmrsqlite"

        ddl_compiler = PMRSQLiteDDLCompiler
        type_compiler = PostgresLikeTypeCompiler
        colspecs = {
            sqltypes.Date: sqlite_base.DATE,
            sqltypes.DateTime: PMRDateTime,
            sqltypes.JSON: sqlite_base.JSON,
            sqltypes.JSON.JSONIndexType: sqlite_base.JSONIndexType,
            sqltypes.JSON.JSONPathType: sqlite_base.JSONPathType,
            sqltypes.Time: sqlite_base.TIME,
        }

        def _json_serializer(self, value):
            return json.dumps(value, sort_keys=True)

    global PostgresLikeSQLitePDialect
    PostgresLikeSQLitePDialect = PostgresLikeDialect
    return PostgresLikeSQLitePDialect


def enable_foreign_key_checks(dbapi_connection, connection_record):
    """Enable foreign key constraint checks.

    SQLite supports FOREIGN KEY syntax when emitting CREATE statements for tables,
    however by default these constraints have no effect on the operation of the table.
    https://docs.sqlalchemy.org/en/latest/dialects/sqlite.html#foreign-key-support
    """
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


@contextlib.contextmanager
def filter_sqlalchemy_warnings(decimal_warnings_enabled=True):
    with warnings.catch_warnings():
        if decimal_warnings_enabled:
            warnings.filterwarnings(
                "ignore",
                (
                    r"^Dialect pmrsqlite\+pysqlite does \*not\* support Decimal objects natively\, and "
                    r"SQLAlchemy must convert from floating point - rounding errors and other issues may occur\. "
                    r"Please consider storing Decimal numbers as strings or integers on this platform for "
                    r"lossless storage\.$"
                ),
                SAWarning,
            )

        yield


def _database_producer():
    i = 1
    while True:
        yield "db{}".format(i)
        i += 1


_database_names = _database_producer()


def create_sqlite_fixture(*ordered_actions, **kwargs):
    """Create a SQLite fixture."""
    scope = kwargs.pop("scope", "function")
    tables = kwargs.pop("tables", None)

    session = kwargs.pop("session", None)
    decimal_warnings = kwargs.pop("decimal_warnings", False)

    postgres_like = kwargs.pop("postgres_like", True)

    if len(kwargs):
        raise KeyError("Unsupported Arguments: {}".format(kwargs))

    dialect_name = "sqlite"

    if postgres_like:
        dialect = make_postgres_like_sqlite_dialect()
        dialect_name = dialect.name
        registry.register("sqlite.{}".format(dialect_name), __name__, "PostgresLikeSQLitePDialect")

    driver_name = "sqlite+{}".format(dialect_name)

    @pytest.fixture(scope=scope)
    def _():
        # XXX: Ideally we eventually make use of the shared memory cache to enable connecting by
        #      credentials with sqlite.
        # database_name = "file:{}?mode=memory&cache=shared".format(next(_database_names))
        database_name = ""

        raw_engine = create_engine("{}:///{}".format(driver_name, database_name))

        # This *must* happen before the connection occurs (implicitly in `EngineManager`).
        event.listen(raw_engine, "connect", enable_foreign_key_checks)

        engine_manager = EngineManager(raw_engine, ordered_actions, tables=tables)
        for engine in engine_manager.manage(session=session):
            with filter_sqlalchemy_warnings(decimal_warnings_enabled=(not decimal_warnings)):
                assign_fixture_credentials(
                    raw_engine,
                    drivername=driver_name,
                    host="",
                    port=None,
                    database=database_name,
                    username="",
                    password="",
                )

                yield engine

        event.remove(raw_engine, "connect", enable_foreign_key_checks)

    return _



from __future__ import absolute_import

import csv
import gzip
import io
import sys

from pytest_mock_resources.compat import boto3
from pytest_mock_resources.patch.redshift.mock_s3_copy import strip


def execute_mock_s3_unload_command(statement, engine):
    params = _parse_s3_command(statement)

    _mock_s3_unload(
        select_statement=params["select_statement"],
        s3_uri=params["s3_uri"],
        aws_secret_access_key=params["aws_secret_access_key"],
        aws_access_key_id=params["aws_access_key_id"],
        engine=engine,
        delimiter=params.get("delimiter", "|"),
        is_gzipped=params["gzip"],
    )


def _parse_s3_command(statement):
    """Format and Parse 'UNLOAD' command."""
    statement = strip(statement)
    params = dict()

    # deleting 'unload'
    tokens = statement.split()[1:]

    # Fetching select statement
    select_statement = ""
    error_flag = False
    for index, token in enumerate(tokens):
        if token.lower() == "to":
            tokens = tokens[index:]
            break
        select_statement += " " + token
    params["select_statement"] = select_statement
    if error_flag:
        raise ValueError(
            (
                "Possibly malformed SELECT Statement. "
                "Statement = {statement}"
                "Redshift fixture only supports S3 Unload statments with the following syntax: "
                "UNLOAD ('select-statement') TO 's3://object-path/name-prefix'"
                "authorization 'aws_access_key_id=<aws_access_key_id>;"
                "aws_secret_access_key=<aws_secret_access_key>'"
                "[GZIP] [DELIMITER [ AS ] 'delimiter-char']"
            ).format(statement=statement)
        )

    # Fetching s3_uri
    if tokens.pop(0).lower() != "to":
        raise ValueError(
            (
                "Possibly malformed S3 URI Format. "
                "Statement = {statement}"
                "Redshift fixture only supports S3 Unload statments with the following syntax: "
                "UNLOAD ('select-statement') TO 's3://object-path/name-prefix'"
                "authorization 'aws_access_key_id=<aws_access_key_id>;"
                "aws_secret_access_key=<aws_secret_access_key>'"
                "[GZIP] [DELIMITER [ AS ] 'delimiter-char']"
            ).format(statement=statement)
        )
    params["s3_uri"] = strip(tokens.pop(0))

    # Fetching authorization
    for token in tokens:
        if "aws_access_key_id" in token.lower() or "aws_secret_access_key" in token.lower():
            # This is because of the following possibiliteis:
            # ... [with ]authorization[ AS] 'aws_access_key_id=x;aws_secret_access_key=y'
            # OR
            # ... [with ]authorization[ AS] 'aws_secret_access_key=y;aws_access_key_id=x'
            # OR
            # ... [with ]authorization[ AS] 'aws_secret_access_key=y;\naws_access_key_id=x'
            # OR
            # ... [with ]authorization[ AS] 'aws_secret_access_key=y; aws_access_key_id=x'
            # Supportred AWS authorization format:
            # [with ]authorization[ AS] 'aws_secret_access_key=y; aws_access_key_id=x'
            # No Support for additional credential formats, eg IAM roles, etc, yet.
            credentials_list = token.split(";")
            for credentials in credentials_list:
                if "aws_access_key_id" in credentials:
                    params["aws_access_key_id"] = credentials.split("=")[-1]
                elif "aws_secret_access_key" in credentials:
                    params["aws_secret_access_key"] = credentials.split("=")[-1]
                else:
                    raise ValueError(
                        (
                            "Possibly malformed AWS Credentials Format. "
                            "Statement = {statement}"
                            "Redshift fixture only supports S3 Copy statments with the following "
                            "syntax: COPY <table_name> FROM [(column 1, [column2, [..]])] '"
                            "<file path on S3 bucket>' "
                            "credentials 'aws_access_key_id=<aws_access_key_id>;"
                            "aws_secret_access_key=<aws_secret_access_key>' "
                            "Supportred AWS credentials format: "
                            "[with ]credentials[ AS] 'aws_secret_access_key=y; aws_access_key_id=x'"
                            " No Support for additional credential formats, eg IAM roles, etc, yet."
                        ).format(statement=statement)
                    )

    # Fetching GZIP Flag
    params["gzip"] = False
    for token in tokens:
        if strip(token.lower()) == "gzip":
            params["gzip"] = True

    # Fetching delimiter
    for index, token in enumerate(tokens):
        if token.lower() == "delimiter":
            try:
                if tokens[index + 1].lower() != "as":
                    params["delimiter"] = strip(tokens[index + 1])
                else:
                    params["delimiter"] = strip(tokens[index + 2])
            except IndexError:
                raise ValueError(
                    (
                        "Possibly malformed Delimiter Format. "
                        "Statement = {statement}"
                        "Redshift fixture only supports S3 Unload statments with the following"
                        "syntax: UNLOAD ('select-statement') TO 's3://object-path/name-prefix'"
                        "authorization 'aws_access_key_id=<aws_access_key_id>;"
                        "aws_secret_access_key=<aws_secret_access_key>'"
                        "[GZIP] [DELIMITER [ AS ] 'delimiter-char']"
                    ).format(statement=statement)
                )
    return params


def _mock_s3_unload(
    select_statement,
    s3_uri,
    aws_secret_access_key,
    aws_access_key_id,
    engine,
    delimiter,
    is_gzipped,
):
    """Execute patched 'unload' command."""
    # Parsing s3 uri
    ending_index = len(s3_uri)
    path_to_file = s3_uri[5:ending_index]
    bucket, key = path_to_file.split("/", 1)

    result = engine.execute(select_statement)
    buffer = get_data_csv(result, is_gzipped=is_gzipped, delimiter=delimiter)

    # Push the data to the S3 Bucket.
    conn = boto3.resource(
        "s3", aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key
    )
    conn.create_bucket(Bucket=bucket)
    obj = conn.Object(bucket, key)
    obj.put(Body=buffer)


def get_data_csv(rows, is_gzipped=False, delimiter="|", **additional_to_csv_options):
    result = io.BytesIO()
    buffer = result
    if is_gzipped:
        buffer = gzip.GzipFile(fileobj=buffer, mode="wb")

    # FUCK you python 2. This is ridiculous!
    wrapper = buffer
    if sys.version_info.major >= 3:
        wrapper = io.TextIOWrapper(buffer)
    else:
        delimiter = delimiter.encode("utf-8")

    writer = csv.DictWriter(
        wrapper,
        fieldnames=rows.keys(),
        delimiter=delimiter,
        quoting=csv.QUOTE_MINIMAL,
        quotechar='"',
        lineterminator="\n",
        skipinitialspace=True,
        doublequote=True,
    )
    writer.writeheader()
    for row in rows:
        writer.writerow(dict(row.items()))

    if sys.version_info.major >= 3:
        wrapper.detach()

    if is_gzipped:
        buffer.close()

    result.seek(0)
    return result



import pytest

from pytest_mock_resources.compat import redis
from pytest_mock_resources.container.redis import RedisConfig
from pytest_mock_resources.fixture.database.generic import assign_fixture_credentials


@pytest.fixture(scope="session")
def pmr_redis_config():
    """Override this fixture with a :class:`RedisConfig` instance to specify different defaults.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_redis_config():
        ...     return RedisConfig(image="redis:6.0")
    """
    return RedisConfig()


def create_redis_fixture(**kwargs):
    """Create a Redis fixture.

    Args:
        scope (str): The scope of the fixture can be specified by the user, defaults to "function".

    Raises:
        KeyError: If any additional arguments are provided to the function than what is necessary.
    """
    scope = kwargs.pop("scope", "function")

    if len(kwargs):
        raise KeyError("Unsupported Arguments: {}".format(kwargs))

    @pytest.fixture(scope=scope)
    def _(_redis_container, pmr_redis_config):
        db = redis.Redis(host=pmr_redis_config.host, port=pmr_redis_config.port)
        db.flushall()
        assign_fixture_credentials(
            db,
            drivername="redis",
            host=pmr_redis_config.host,
            port=pmr_redis_config.port,
            database=None,
            username=None,
            password=None,
        )
        return db

    return _



# flake8: noqa
from pytest_mock_resources.fixture.database.mongo import create_mongo_fixture, pmr_mongo_config
from pytest_mock_resources.fixture.database.redis import create_redis_fixture, pmr_redis_config
from pytest_mock_resources.fixture.database.relational import (
    create_mysql_fixture,
    create_postgres_fixture,
    create_redshift_fixture,
    create_sqlite_fixture,
    pmr_mysql_config,
    pmr_postgres_config,
    Rows,
    Statements,
)



# flake8: noqa
from pytest_mock_resources.fixture.database.relational.generic import Rows, Statements
from pytest_mock_resources.fixture.database.relational.mysql import (
    create_mysql_fixture,
    pmr_mysql_config,
)
from pytest_mock_resources.fixture.database.relational.postgresql import (
    create_postgres_fixture,
    pmr_postgres_config,
)
from pytest_mock_resources.fixture.database.relational.redshift import create_redshift_fixture
from pytest_mock_resources.fixture.database.relational.sqlite import create_sqlite_fixture



import enum

from pytest_mock_resources.fixture.database import Statements


@enum.unique
class UdfLanguage(enum.Enum):
    PLPGSQL = "plpgsql"
    PLPYTHON = "plpythonu"
    SQL = "SQL"


def create_udf(name, args, returns, body, language, schema="public"):
    _template = """
    CREATE FUNCTION {schema}.{name} ({args}) RETURNS {returns} AS $${body}$$ LANGUAGE {language};
    """

    return _template.format(
        schema=schema, name=name, args=args, returns=returns, body=body, language=language
    )


dateadd_kwargs = dict(body="SELECT d + (n::VARCHAR || i)::INTERVAL", language=UdfLanguage.SQL.value)

dateadd_date = create_udf(
    name="DATEADD",
    args="i VARCHAR, n INTEGER, d DATE",
    returns="TIMESTAMP WITHOUT TIME ZONE",
    **dateadd_kwargs
)

dateadd_timestamp = create_udf(
    name="DATEADD",
    args="i VARCHAR, n INTEGER, d TIMESTAMP WITHOUT TIME ZONE",
    returns="TIMESTAMP WITHOUT TIME ZONE",
    **dateadd_kwargs
)

dateadd_timestamptz = create_udf(
    name="DATEADD",
    args="i VARCHAR, n INTEGER, d TIMESTAMP WITH TIME ZONE",
    returns="TIMESTAMP WITH TIME ZONE",
    **dateadd_kwargs
)

date_add_date = create_udf(
    name="DATE_ADD",
    args="i VARCHAR, n INTEGER, d DATE",
    returns="TIMESTAMP WITHOUT TIME ZONE",
    **dateadd_kwargs
)

date_add_timestamp = create_udf(
    name="DATE_ADD",
    args="i VARCHAR, n INTEGER, d TIMESTAMP WITHOUT TIME ZONE",
    returns="TIMESTAMP WITHOUT TIME ZONE",
    **dateadd_kwargs
)

date_add_timestamptz = create_udf(
    name="DATE_ADD",
    args="i VARCHAR, n INTEGER, d TIMESTAMP WITH TIME ZONE",
    returns="TIMESTAMP WITH TIME ZONE",
    **dateadd_kwargs
)


datediff_kwargs = dict(
    returns="BIGINT",
    # Credit: https://gist.github.com/JoshuaGross/18b9bb1db8021efc88884cbd8dc8fddb
    body="""
       DECLARE
         diff_interval INTERVAL;
         diff INT = 0;
         years_diff INT = 0;

       BEGIN
         IF units NOT IN (
         'y', 'yr', 'yrs', 'year', 'years',
         'month', 'months', 'mon', 'mons',
         'week', 'weeks', 'w',
         'day', 'days', 'd',
         'hour', 'hours', 'h', 'hr', 'hrs',
         'minute', 'minutes', 'm', 'min', 'mins',
         'second', 'seconds'
         ) THEN
           RAISE EXCEPTION 'Invalid unit %% specified', units;
         END IF;

         IF units IN (
         'y', 'yr', 'yrs', 'year', 'years',
         'month', 'months', 'mon', 'mons'
         ) THEN
           years_diff = DATE_PART('year', end_t) - DATE_PART('year', start_t);

           IF units IN ('y', 'yr', 'yrs', 'year', 'years') THEN
             RETURN years_diff::BIGINT;
           ELSE
             RETURN (
             years_diff * 12 + (DATE_PART('month', end_t) - DATE_PART('month', start_t))
             )::BIGINT;
           END IF;
         END IF;

         IF pg_typeof(start_t) = pg_typeof(DATE('2000-01-01')) THEN
           diff_interval = (end_t - start_t) * '1 day'::INTERVAL;
         ELSE
           diff_interval = (end_t - start_t);
         END IF;

         diff = diff + DATE_PART('day', diff_interval);

         IF units IN ('week', 'weeks', 'w') THEN
           diff = diff/7;
           RETURN diff::BIGINT;
         END IF;

         IF units IN ('day', 'days', 'd') THEN
           RETURN diff::BIGINT;
         END IF;

         diff = diff * 24 + DATE_PART('hour', diff_interval);

         IF units IN ('hour', 'hours', 'h', 'hr', 'hrs') THEN
            RETURN diff::BIGINT;
         END IF;

         diff = diff * 60 + DATE_PART('minute', diff_interval);

         IF units IN ('minute', 'minutes', 'm', 'min', 'mins') THEN
            RETURN diff::BIGINT;
         END IF;

         diff = diff * 60 + DATE_PART('second', diff_interval);

         RETURN diff::BIGINT;

       END;
    """,
    language=UdfLanguage.PLPGSQL.value,
)

datediff_date = create_udf(
    name="DATEDIFF", args="units VARCHAR, start_t DATE, end_t DATE", **datediff_kwargs
)

datediff_timestamp = create_udf(
    name="DATEDIFF", args="units VARCHAR, start_t TIMESTAMP, end_t TIMESTAMP", **datediff_kwargs
)

datediff_timestamptz = create_udf(
    name="DATEDIFF",
    args="units VARCHAR, start_t TIMESTAMP WITH TIME ZONE, end_t TIMESTAMP WITH TIME ZONE",
    **datediff_kwargs
)

REDSHIFT_UDFS = Statements(
    dateadd_date,
    dateadd_timestamp,
    dateadd_timestamptz,
    date_add_date,
    date_add_timestamp,
    date_add_timestamptz,
    datediff_date,
    datediff_timestamp,
    datediff_timestamptz,
)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pytest_mock_resources',
 'pytest_mock_resources.container',
 'pytest_mock_resources.fixture',
 'pytest_mock_resources.fixture.database',
 'pytest_mock_resources.fixture.database.relational',
 'pytest_mock_resources.fixture.database.relational.redshift',
 'pytest_mock_resources.patch',
 'pytest_mock_resources.patch.redshift']

package_data = \
{'': ['*']}

install_requires = \
['docker>=2.5.1', 'pytest>=1.0', 'responses', 'sqlalchemy']

extras_require = \
{':python_version >= "2.7" and python_version < "2.8"': ['functools32'],
 'mongo': ['pymongo'],
 'mysql': ['pymysql>=1.0'],
 'postgres': ['psycopg2'],
 'postgres-binary': ['psycopg2-binary'],
 'redis': ['redis'],
 'redshift': ['moto', 'boto3', 'sqlparse']}

entry_points = \
{'console_scripts': ['pmr = pytest_mock_resources.cli:main'],
 'pytest11': ['pytest_mock_resources = pytest_mock_resources']}

setup_kwargs = {
    'name': 'pytest-mock-resources',
    'version': '2.0.0',
    'description': 'A pytest plugin for easily instantiating reproducible mock resources.',
    'long_description': '![CircleCI](https://img.shields.io/circleci/build/gh/schireson/pytest-mock-resources/master) [![codecov](https://codecov.io/gh/schireson/pytest-mock-resources/branch/master/graph/badge.svg)](https://codecov.io/gh/schireson/pytest-mock-resources) [![Documentation Status](https://readthedocs.org/projects/pytest-mock-resources/badge/?version=latest)](https://pytest-mock-resources.readthedocs.io/en/latest/?badge=latest)\n\n\n## Introduction\n\nCode which depends on external resources such a databases (postgres, redshift, etc) can be difficult\nto write automated tests for. Conventional wisdom might be to mock or stub out the actual database\ncalls and assert that the code works correctly before/after the calls.\n\nHowever take the following, _simple_ example:\n\n```python\ndef serialize(users):\n    return [\n        {\n            \'user\': user.serialize(),\n            \'address\': user.address.serialize(),\n            \'purchases\': [p.serialize() for p in user.purchases],\n        }\n        for user in users\n    ]\n\ndef view_function(session):\n    users = session.query(User).join(Address).options(selectinload(User.purchases)).all()\n    return serialize(users)\n```\n\nSure, you can test `serialize`, but whether the actual **query** did the correct thing _truly_\nrequires that you execute the query.\n\n## The Pitch\n\nHaving tests depend upon a **real** postgres instance running somewhere is a pain, very fragile,\nand prone to issues across machines and test failures.\n\nTherefore `pytest-mock-resources` (primarily) works by managing the lifecycle of docker containers\nand providing access to them inside your tests.\n\nAs such, this package makes 2 primary assumptions:\n* You\'re using `pytest` (hopefully that\'s appropriate, given the package name)\n* For many resources, `docker` is required to be available and running (or accessible through remote docker).\n\nIf you aren\'t familiar with Pytest Fixtures, you can read up on them in the [Pytest documentation](https://docs.pytest.org/en/latest/fixture.html).\n\nIn the above example, your test file could look something like\n\n```python\nfrom pytest_mock_resources import create_postgres_fixture\nfrom models import ModelBase\n\npg = create_postgres_fixture(ModelBase, session=True)\n\ndef test_view_function_empty_db(pg):\n  response = view_function(pg)\n  assert response == ...\n\ndef test_view_function_user_without_purchases(pg):\n  pg.add(User(...))\n  pg.flush()\n\n  response = view_function(pg)\n  assert response == ...\n\ndef test_view_function_user_with_purchases(pg):\n  pg.add(User(..., purchases=[Purchase(...)]))\n  pg.flush()\n\n  response = view_function(pg)\n  assert response == ...\n```\n\n## Existing Resources (many more possible)\n\n* SQLite\n\n  ```python\n  from pytest_mock_resources import create_sqlite_fixture\n  ```\n\n* Postgres\n\n  ```python\n  from pytest_mock_resources import create_postgres_fixture\n  ```\n\n* Redshift\n\n  **note** Uses postgres under the hood, but the fixture tries to support as much\n  redshift functionality as possible (including redshift\'s `COPY`/`UNLOAD` commands).\n\n  ```python\n  from pytest_mock_resources import create_redshift_fixture\n  ```\n\n* Mongo\n\n  ```python\n  from pytest_mock_resources import create_mongo_fixture\n  ```\n\n* Redis\n\n  ```python\n  from pytest_mock_resources import create_redis_fixture\n  ```\n  \n * MySQL\n \n   ```python\n   from pytest_mock_resources import create_mysql_fixture\n   ```\n\n## Installing\n\n```bash\n# Basic fixture support\npip install "pytest-mock-resources"\n\n# For postgres install EITHER of the following:\npip install "pytest-mock-resources[postgres-binary]"\npip install "pytest-mock-resources[postgres]"\n\n# For redshift install EITHER of the following:\n# (redshift fixtures require postgres dependencies...)\npip install "pytest-mock-resources[postgres, redshift]"\npip install "pytest-mock-resources[postgres-binary, redshift]"\n\n# For mongo install the following:\npip install "pytest-mock-resources[mongo]"\n\n# For redis\npip install "pytest-mock-resources[redis]"\n\n# For mysql\npip install "pytest-mock-resources[mysql]"\n```\n\n## Possible Future Resources\n\n* Rabbit Broker\n* AWS Presto\n\nFeel free to file an [issue](https://github.com/schireson/pytest-mock-resources/issues) if you find any bugs or want to start a conversation around a mock resource you want implemented!\n\n\n## Python 2\nReleases in the 1.x series were supportive of python 2. However starting from 2.0.0,\nsupport for python 2 was dropped. We may accept bugfix PRs for the 1.x series,\nhowever new development and features will not be backported.\n',
    'author': 'Omar Khan',
    'author_email': 'oakhan3@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/schireson/pytest-mock-resources',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4',
}


setup(**setup_kwargs)



from sqlalchemy.engine.url import URL


class Credentials:
    """Return as `pmr_credentials` attribute on supported docker-based fixtures.

    Examples:
        It's also directly dict-able.
        >>> creds = Credentials('d', 'l', 'p', 'baz', 'user', 'pass')
        >>> dict_creds = dict(creds)
    """

    def __init__(self, drivername, host, port, database, username, password):
        self.drivername = drivername
        self.host = host
        self.port = port
        self.database = database
        self.username = username
        self.password = password

    def __iter__(self):
        for item in self.__dict__:
            yield (item, self[item])

    def __getitem__(self, item):
        return self.__dict__[item]

    def as_url(self):
        """Return a stringified dbapi URL string."""
        return str(self.as_sqlalchemy_url())

    def as_sqlalchemy_url(self):
        """Return a sqlalchemy :class:`sqlalchemy.engine.url.URL`."""
        return URL(
            drivername=self.drivername,
            host=self.host,
            port=self.port,
            database=self.database,
            username=self.username,
            password=self.password,
        )

    def as_sqlalchemy_url_kwargs(self):
        """Return the valid arguments to sqlalchemy :class:`sqlalchemy.engine.url.URL`."""
        return dict(self)

    def as_psycopg2_kwargs(self):
        """Return the valid arguments to sqlalchemy :class:`psycopg2.connect`."""
        return {
            "host": self.host,
            "port": self.port,
            "user": self.username,
            "password": self.password,
            "dbname": self.database,
        }

    def as_mongo_kwargs(self):
        """Return the valid arguments to a mongo client."""
        return {
            "host": self.host,
            "port": self.port,
            "username": self.username,
            "password": self.password,
            "authSource": self.database,
        }

    def as_redis_kwargs(self):
        """Return the valid arguments to a redis client."""
        return {
            "host": self.host,
            "port": self.port,
            "db": self.database,
            "username": self.username,
            "password": self.password,
        }


def assign_fixture_credentials(engine, **credentials):
    engine.pmr_credentials = Credentials(**credentials)



from __future__ import absolute_import

from sqlalchemy.sql.elements import TextClause
from sqlalchemy.sql.expression import Insert, Select, Update

from pytest_mock_resources.compat import sqlparse
from pytest_mock_resources.patch.redshift.mock_s3_copy import execute_mock_s3_copy_command, strip
from pytest_mock_resources.patch.redshift.mock_s3_unload import execute_mock_s3_unload_command

SQLALCHEMY_BASES = (Select, Insert, Update, TextClause)


def substitute_execute_with_custom_execute(engine):
    """Substitute the default execute method with a custom execute for copy and unload command."""
    default_execute = engine.execute

    def custom_execute(statement, *args, **kwargs):
        if not isinstance(statement, SQLALCHEMY_BASES) and strip(statement).lower().startswith(
            "copy"
        ):
            return execute_mock_s3_copy_command(statement, engine)
        if not isinstance(statement, SQLALCHEMY_BASES) and strip(statement).lower().startswith(
            "unload"
        ):
            return execute_mock_s3_unload_command(statement, engine)
        return default_execute(statement, *args, **kwargs)

    def handle_multiple_statements(statement, *args, **kwargs):
        """Split statement into individual sql statements and execute.

        Splits multiple statements by ';' and executes each.
        NOTE: Only the result of the last statements is returned.
        """
        statements_list = parse_multiple_statements(statement)
        result = None
        for statement in statements_list:
            result = custom_execute(statement, *args, **kwargs)

        return result

    # Now each statement is handled as if it contains multiple sql statements
    engine.execute = handle_multiple_statements
    return engine


def parse_multiple_statements(statement):
    """Split the given sql statement into a list of individual sql statements."""
    statements_list = []

    # Ignore SQLAlchemy Text Objects.
    if isinstance(statement, SQLALCHEMY_BASES):
        statements_list.append(statement)
        return statements_list

    # Prprocess input statement
    statement = _preprocess(statement)

    statements_list = [str(statement) for statement in sqlparse.split(statement)]

    return statements_list


def _preprocess(statement):
    """Preprocess the input statement."""
    statement = statement.strip()
    # Replace any occourance of " with '.
    statement = statement.replace('"', "'")
    if statement[-1] != ";":
        statement += ";"
    return statement



class ImportAdaptor(object):
    __wrapped__ = False

    def __init__(self, package, recommended_extra, fail_message=None, **attrs):
        self.package = package
        self.recommended_extra = recommended_extra
        self.fail_message = fail_message

        for key, value in attrs.items():
            setattr(self, key, value)

    def fail(self):
        if self.fail_message:
            fail_message = self.fail_message
        else:
            fail_message = "Cannot use {recommended_extra} fixtures without {package}. pip install pytest-mock-resources[{recommended_extra}]".format(
                package=self.package, recommended_extra=self.recommended_extra
            )

        raise RuntimeError(fail_message)

    def __getattr__(self, attr):
        self.fail()


try:
    from unittest import mock
except ImportError:
    import mock  # type: ignore # noqa

try:
    import functools32 as functools
except ImportError:
    import functools  # type: ignore # noqa

try:
    import psycopg2
except ImportError:
    fail_message = (
        "Cannot use postgres/redshift fixtures without psycopg2.\n"
        "pip install pytest-mock-resources[postgres] or pytest-mock-resources[[postgres-binary].\n"
        "Additionally, pip install pytest-mock-resources[redshift] for redshift fixtures."
    )
    psycopg2 = ImportAdaptor(
        "psycopg2",
        "postgres",
        fail_message=fail_message,
        extensions=ImportAdaptor(
            "psycopg2", "psycopg2", fail_message=fail_message, cursor=ImportAdaptor
        ),
    )

try:
    import boto3
except ImportError:
    boto3 = ImportAdaptor("boto3", "redshift")

try:
    import moto
except ImportError:
    moto = ImportAdaptor("moto", "redshift")

try:
    import sqlparse
except ImportError:
    sqlparse = ImportAdaptor(
        "sqlparse",
        "redshift",
    )

try:
    import pymongo
except ImportError:
    pymongo = ImportAdaptor("pymongo", "mongo")

try:
    import redis
except ImportError:
    redis = ImportAdaptor("redis", "redis")  # type: ignore

try:
    import pymysql
except ImportError:
    pymysql = ImportAdaptor("pymysql", "mysql")
